--!parent
--!encoding=utf-8

--****************************************************************************
-- @file QDiskImager - main class
-- @author Valentin Schmidt
-- @version 0.5
--****************************************************************************

global $

-- static
property APP_NAME
property APP_VERSION
property EXIT_CODE_ABORT

----------------------------------------
-- widgets
----------------------------------------
property pMainWindow
property pStatusBar
property pProgressBar
property pTabWidget

-- Create Image
property pGroupBoxCreateSource
property pGroupBoxCreateTarget

property pComboBoxCreateDisk
property pCheckBoxCreateOnlyPartitions
property pLineEditCreateFile
property pPushButtonCreateStart
property pPushButtonCreateCancel
property pPlainTextEditCreateDiskInfo
property pLineEditCreateHash
property pComboBoxCreateHash
property pPushButtonCreateHash

-- Restore Image
property pGroupBoxRestoreSource
property pGroupBoxRestoreTarget

property pLineEditRestoreFile
property pCheckBoxRestoreOnlyPartitions
property pComboBoxRestoreDisk
property pPushButtonRestoreStart
property pPushButtonRestoreCancel
property pPlainTextEditRestoreDiskInfo
property pPlainTextEditRestoreImageInfo

property pLineEditRestoreHash
property pComboBoxRestoreHash
property pPushButtonRestoreHash

-- Mount
property pTableWidgetMount
property pPushButtonMount
property pPushButtonUnmount
property pPushButtonUnmountAll

-- Linux Reader
property pComboBoxLinuxPartition
property pTableWidgetLinux

-- Objects
property pState
property pActions
property pTrayIcon

-- Vars

-- Create
property pCreateSectorSize
property pCreateLastSector
property pCreateDiskSize
property pCreateImgSize
property pCreateDiskIsGPT

-- Restore
property pRestoreSectorSize
property pRestoreLastSector
property pRestoreDiskSize
property pRestoreImgSize
property pRestoreImageIsGPT

property pIsWindows8Point1OrGreater
property pLastDir
property pDiskInfos
property pZeroMounts
property pTmpZero

----------------------------------------
-- @constructor
----------------------------------------
on new (me)
	--_player.debugPlayBackEnabled = TRUE

	-- libs
	$.import("binfile")
	$.import("crypto")
	$.import("qt")
	$.import("shell")
	$.import("string")
	$.import("wfx")
	
	me.APP_NAME = "QDiskImager"
	me.APP_VERSION = 5

	me.EXIT_CODE_ABORT = -1073741510

	_movie.stage.title = me.APP_NAME

	v = getRealOSVersion() -- [6, 3]
	me.pIsWindows8Point1OrGreater = (v[1]>6 or (v[1]=6 and v[2]>=3))

	-- prepend bin dir to path
	setEnvVar("PATH", $.PATH&"bin;"&getEnvVar("PATH"))

	me.pState = $.Qt.Settings.new($.Qt.QSettings.NativeFormat, "fx", APP_NAME)

	$.Qt.Resource.registerResourceFile($.PATH&"main.rcc")

	-- create main window
	uiLoader = $.Qt.UiLoader.new()

	me.pMainWindow = uiLoader.load($.PATH&"main.ui")
	me.pMainWindow.setWindowTitle(me.APP_NAME)
	me.pMainWindow.connect($.Qt.Event.Close, me, #slotQuit, TRUE)

	me.pStatusBar = me.pMainWindow.findChild("statusbar", #StatusBar)

	me.pProgressBar = $.Qt.ProgressBar.new(me.pMainWindow)
	me.pProgressBar.hide()
	me.pProgressBar.setMinimum(0)
	me.pProgressBar.setMaximum(100)
	me.pStatusBar.addPermanentWidget(me.pProgressBar)

	me.pTabWidget = me.pMainWindow.findChild("tabWidget", #TabWidget)

	----------------------------------------
	-- Create Image
	----------------------------------------
	me.pGroupBoxCreateSource = me.pMainWindow.findChild("groupBoxCreateSource", #GroupBox)
	me.pGroupBoxCreateTarget = me.pMainWindow.findChild("groupBoxCreateTarget", #GroupBox)

	me.pComboBoxCreateDisk = me.pMainWindow.findChild("comboBoxCreateDisk", #ComboBox)
	me.pComboBoxCreateDisk.connect(#activatedIndex, me, #slotCreateShowDiskInfo)

	me.pCheckBoxCreateOnlyPartitions = me.pMainWindow.findChild("checkBoxCreateOnlyPartitions", #CheckBox)

	btn = me.pMainWindow.findChild("toolButtonCreateDiskRefresh", #ToolButton)
	btn.connect(#clicked, me, #slotUpdateDisks)

	me.pLineEditCreateFile = me.pMainWindow.findChild("lineEditCreateFile", #LineEdit)

	btn = me.pMainWindow.findChild("pushButtonCreateFile", #PushButton)
	btn.connect(#clicked, me, #slotCreateFileSelect)

	me.pPushButtonCreateStart = me.pMainWindow.findChild("pushButtonCreateStart", #PushButton)
	me.pPushButtonCreateStart.connect(#clicked, me, #slotCreateStart)

	me.pPushButtonCreateCancel = me.pMainWindow.findChild("pushButtonCreateCancel", #PushButton)
	me.pPushButtonCreateCancel.connect(#clicked, me, #slotCreateCancel)

	me.pPlainTextEditCreateDiskInfo = me.pMainWindow.findChild("plainTextEditCreateDiskInfo", #PlainTextEdit)

	me.pLineEditCreateHash = me.pMainWindow.findChild("lineEditCreateHash", #LineEdit)
	me.pComboBoxCreateHash = me.pMainWindow.findChild("comboBoxCreateHash", #ComboBox)
	me.pPushButtonCreateHash = me.pMainWindow.findChild("pushButtonCreateHash", #PushButton)
	me.pPushButtonCreateHash.connect(#clicked, me, #slotCreateGenerateHash)

	----------------------------------------
	-- Restore Image
	----------------------------------------
	me.pGroupBoxRestoreSource = me.pMainWindow.findChild("groupBoxRestoreSource", #GroupBox)
	me.pGroupBoxRestoreTarget = me.pMainWindow.findChild("groupBoxRestoreTarget", #GroupBox)

	me.pLineEditRestoreFile = me.pMainWindow.findChild("lineEditRestoreFile", #LineEdit)

	btn = me.pMainWindow.findChild("pushButtonRestoreFile", #PushButton)
	btn.connect(#clicked, me, #slotRestoreFileSelect)

	me.pCheckBoxRestoreOnlyPartitions = me.pMainWindow.findChild("checkBoxRestoreOnlyPartitions", #CheckBox)

	me.pComboBoxRestoreDisk = me.pMainWindow.findChild("comboBoxRestoreDisk", #ComboBox)
	me.pComboBoxRestoreDisk.connect(#activatedIndex, me, #slotRestoreShowDiskInfo)

	btn = me.pMainWindow.findChild("toolButtonRestoreDiskRefresh", #ToolButton)
	btn.connect(#clicked, me, #slotUpdateDisks)

	me.pPushButtonRestoreStart = me.pMainWindow.findChild("pushButtonRestoreStart", #PushButton)
	me.pPushButtonRestoreStart.connect(#clicked, me, #slotRestoreStart)

	me.pPushButtonRestoreCancel = me.pMainWindow.findChild("pushButtonRestoreCancel", #PushButton)
	me.pPushButtonRestoreCancel.connect(#clicked, me, #slotRestoreCancel)

	me.pPlainTextEditRestoreDiskInfo = me.pMainWindow.findChild("plainTextEditRestoreDiskInfo", #PlainTextEdit)
	me.pPlainTextEditRestoreImageInfo = me.pMainWindow.findChild("plainTextEditRestoreImageInfo", #PlainTextEdit)

	me.pLineEditRestoreHash = me.pMainWindow.findChild("lineEditRestoreHash", #LineEdit)
	me.pComboBoxRestoreHash = me.pMainWindow.findChild("comboBoxRestoreHash", #ComboBox)
	me.pPushButtonRestoreHash = me.pMainWindow.findChild("pushButtonRestoreHash", #PushButton)
	me.pPushButtonRestoreHash.connect(#clicked, me, #slotRestoreGenerateHash)

	----------------------------------------
	-- mount
	----------------------------------------
	me.pTableWidgetMount = me.pMainWindow.findChild("tableWidgetMount", #TableWidget)
	me.pTableWidgetMount.setColumnWidth(1, 638-200)
	me.pTableWidgetMount.setHorizontalHeaderDefaultAlignment($.Qt.AlignLeft)
	me.pTableWidgetMount.connect(#currentItemChanged , me, #slotMountTableWidgetCurrentItemChanged)
	me.pTableWidgetMount.connect(#cellDoubleClicked , me, #slotMountTableWidgetCellDoubleClicked)

	me.pPushButtonMount = me.pMainWindow.findChild("pushButtonMount", #PushButton)
	me.pPushButtonMount.connect(#clicked, me, #slotMountNew)
	me.pPushButtonUnmount = me.pMainWindow.findChild("pushButtonUnmount", #PushButton)
	me.pPushButtonUnmount.connect(#clicked, me, #slotMountUnmount)
	me.pPushButtonUnmountAll = me.pMainWindow.findChild("pushButtonUnmountAll", #PushButton)
	me.pPushButtonUnmountAll.connect(#clicked, me, #slotMountUnmountAll)

	----------------------------------------
	-- LinuxReader
	----------------------------------------
	me.pComboBoxLinuxPartition = me.pMainWindow.findChild("comboBoxLinuxPartition", #ComboBox)

	btn = me.pMainWindow.findChild("toolButtonLinuxPartitionRefresh", #ToolButton)
	btn.connect(#clicked, me, #slotLinuxUpdatePartitions)

	me.pTableWidgetLinux = me.pMainWindow.findChild("tableWidgetLinux", #TableWidget)
	me.pTableWidgetLinux.setColumnWidth(0, 638-100)	
	me.pTableWidgetLinux.setHorizontalHeaderDefaultAlignment($.Qt.AlignLeft)
	me.pTableWidgetLinux.connect(#cellDoubleClicked, me, #slotTableWidgetLinuxCellDoubleClicked)
	me.pTableWidgetLinux.sortItems(0, 0)
	me.pTableWidgetLinux.setContextMenuPolicy($.Qt.CustomContextMenu)
	me.pTableWidgetLinux.connect(#customContextMenuRequested, me, #slotTableWidgetLinuxContextMenuRequested)
	me.pTableWidgetLinux[#contextMenu] = $.Qt.Menu.new(me.pMainWindow)

	action = $.Qt.Action.new(me.pTableWidgetLinux[#contextMenu])
	action.setText("Save Selected")
	action.connect(#triggered, me, #slotTableWidgetLinuxSave)
	me.pTableWidgetLinux[#contextMenu].addAction(action)

	$.wfx.FsSetProgressCallback(#slotLinuxGetFileProgress, me)

	----------------------------------------
	-- menu actions
	----------------------------------------
	me.pActions = [:]
	menuBar = me.pMainWindow.findChild("menubar", #MenuBar)
	menus = menuBar.findChildren(VOID, #Menu)
	repeat with m in menus
		actions = m.actions()
		repeat with a in actions
			str = a.objectName()
			if str<>"" then me.pActions[str] = a
		end repeat
	end repeat

	me.pActions.actionAbout.connect(#triggered, me, #slotAbout)

	-- drop support
	--> No drag&drop allowed in elevated app!
	--me.pMainWindow.setAcceptDrops(TRUE)
	--$.Qt.registerDropHandler(me.pMainWindow, #slotRestoreSourceFilesDropped, me)

	me.pLastDir = me.pState.Value("LastDir")
	if voidP(me.pLastDir) then me.pLastDir = "C:\"

	me.pCreateSectorSize = 512
	me.pCreateLastSector = 0
	me.pCreateDiskSize = 0
	me.pCreateImgSize = 0
	me.pCreateDiskIsGPT = FALSE

	me.pRestoreSectorSize = 512
	me.pRestoreLastSector = 0
	me.pRestoreDiskSize = 0
	me.pRestoreImgSize = 0
	me.pRestoreImageIsGPT = FALSE

	-- tray icon
	me.pTrayIcon = $.Qt.SystemTrayIcon.new(me.pMainWindow)
	me.pTrayIcon.setIcon(":/main/icons/app.png")
	me.pTrayIcon.connect(#activated, me, #slotTrayIconActivated)
	--me.pTrayIcon.show()

	me.pMainWindow.connect($.Qt.Event.WindowStateChange, me, #slotWindowStateChanged)

	-- load Disks
	me.slotUpdateDisks()

	-- load Linux Partitions
	me.slotLinuxUpdatePartitions()

	-- restore window state
	val = me.pState.value("MainWindow/Geometry")
	if not voidP(val) then me.pMainWindow.restoreGeometry(val)
	val = me.pState.value("MainWindow/State")
	if not voidP(val) then me.pMainWindow.restoreState(val)

	me.pMainWindow.show()
	
	hwnd = me.pMainWindow.winId()
	taskBarProgressInit(hwnd)

	-- assign alertHook to capture script errors
	if not _player.debugPlayBackEnabled then _player.alertHook = me

	return me
end

--------------------------------------
--
--------------------------------------
on alertHook (me, aMsg, aText, aType)
	case (aType) of
	#alert: return 0
	otherwise:
		me.pStatusBar.showMessage(aMsg&": "&aText)
		return 1
	end case
end

on __UTILS
end

--------------------------------------
--
--------------------------------------
on _formatFileSize (me, b)
	fp = the floatPrecision
	the floatPrecision = 2
	if b>power(1024, 4) then
		res = b/power(1024, 4) && "TB"
	else if b>power(1024, 3) then
		res = b/power(1024, 3) && "GB"
	else if b>power(1024, 2) then
		res = b/power(1024, 2) && "MB"
	else if b>1024 then
		res = b/1024. && "KB"
	else
		res = b && "Bytes"
	end if
	the floatPrecision = fp
	return res
end

--------------------------------------
--
--------------------------------------
on _uiCreateSetEnabled (me, flag)
	if flag then
		me.pTabWidget.setTabEnabled(1, TRUE)

		me.pGroupBoxCreateSource.setDisabled(FALSE)
		me.pGroupBoxCreateTarget.setDisabled(FALSE)
		me.pPushButtonCreateStart.setDisabled(FALSE)

		me.pPushButtonCreateCancel.setDisabled(TRUE)

		me.pProgressBar.hide()
	else
		me.pTabWidget.setTabEnabled(1, FALSE)

		me.pGroupBoxCreateSource.setDisabled(TRUE)
		me.pGroupBoxCreateTarget.setDisabled(TRUE)
		me.pPushButtonCreateStart.setDisabled(TRUE)

		me.pPushButtonCreateCancel.setDisabled(FALSE)
	end if
end

--------------------------------------
--
--------------------------------------
on _uiRestoreSetEnabled (me, flag)
	if flag then
		me.pTabWidget.setTabEnabled(0, TRUE)

		me.pGroupBoxRestoreSource.setDisabled(FALSE)
		me.pGroupBoxRestoreTarget.setDisabled(FALSE)
		me.pPushButtonRestoreStart.setDisabled(FALSE)

		me.pPushButtonRestoreCancel.setDisabled(TRUE)

		me.pProgressBar.hide()
	else
		me.pTabWidget.setTabEnabled(0, FALSE)

		me.pGroupBoxRestoreSource.setDisabled(TRUE)
		me.pGroupBoxRestoreTarget.setDisabled(TRUE)
		me.pPushButtonRestoreStart.setDisabled(TRUE)

		me.pPushButtonRestoreCancel.setDisabled(FALSE)
	end if
end

--------------------------------------
--
--------------------------------------
on _int (me, f)
	fp = the floatPrecision
	the floatPrecision = 0
	res = string(f)
	the floatPrecision = fp
	return res
end

--------------------------------------
--
--------------------------------------
on _zeroMemory (me, disk, mountNum)
	me.pZeroMounts = me.pDiskInfos[disk].mounts.duplicate()
	me._zeroNextMount(0)
end

--------------------------------------
--
--------------------------------------
on _zeroNextMount (me, exitCode)
	if exitCode=EXIT_CODE_ABORT then
		me.pStatusBar.showMessage("Running task was aborted by user.")
		-- clean up
		$.shell.shell_cmd("del "&me.pTmpZero)
		return
	end if

	if me.pZeroMounts.count=0 then return me.slotCreateStart(TRUE)
	me.pStatusBar.showMessage("Zeroing free space on volume "&me.pZeroMounts[1]&"...")

	me.pTmpZero = me.pZeroMounts[1]&"___zero___"&random(the maxInteger)
	cmd = "cat /dev/zero>"&me.pTmpZero&"&del "&me.pTmpZero

	me.pZeroMounts.deleteAt(1)
	props = [:]
	props["cb_complete"] = #_zeroNextMount
	props["cb_target"] = me
	props["timeout"] = 0
	$.shell.shell_cmd_thread(cmd, props)
end

--------------------------------------
--
--------------------------------------
on _loadLinuxPath (me, p)
	rows = []
	res = $.wfx.FsFindFirst(p)
	repeat while (res<>0)
		if res["FileName"]<>".." then rows.add(res)
		res = $.wfx.FsFindNext()
	end repeat
	$.wfx.FsFindClose()

	-- show in table
	me.pTableWidgetLinux.setSortingEnabled(FALSE)	
	me.pTableWidgetLinux.clearContents()
	
	row = 0

	_player.itemDelimiter = "\"
	if p.item.count>3 then
		me.pTableWidgetLinux.setRowCount(rows.count+1)
		tItem = $.Qt.TableWidgetItem.new()
		tItem.setText("..")
		tItem.setIcon($.PATH, TRUE)
	    tItem.setData($.Qt.UserRole, p.item[1..p.item.count-2]&"\") -- TYPE_FOLDER
	    tItem.setSortValue(row)
	    me.pTableWidgetLinux.setItem(0, 0, tItem)
	    tItem = $.Qt.TableWidgetItem.new()
	    tItem.setText("")
	    me.pTableWidgetLinux.setItem(0, 1, tItem)
	    row = 1
	else
		me.pTableWidgetLinux.setRowCount(rows.count)
	end if

	repeat with res in rows
		tItem = $.Qt.TableWidgetItem.new()
		tItem.setText(res["FileName"])

		is_dir = (res["FileAttributes"] contains "D")
		if is_dir then
			tItem.setIcon($.PATH, TRUE)
			tItem.setData($.Qt.UserRole, p&res["FileName"]&"\")
			tItem.setSortValue(row)
		else
			tItem.setIcon("", TRUE)
			tItem.setData($.Qt.UserRole, p&res["FileName"])
			tItem.setSortValue(1000+row)
		end if

		me.pTableWidgetLinux.setItem(row, 0, tItem)

		tItem = $.Qt.TableWidgetItem.new()
		tItem.setSortValue(res["FileSize"])
		if is_dir then
			tItem.setText("")
		else
			tItem.setText(me._formatFileSize(string(res["FileSize"])))
		end if

		me.pTableWidgetLinux.setItem(row, 1, tItem)

		row = row + 1
	end repeat

	me.pTableWidgetLinux.setSortingEnabled(TRUE)
end

--------------------------------------
--
--------------------------------------
on _saveLinuxPath (me, remotePath, localDir)
	--put "_saveLinuxPath", remotePath, localDir
	is_dir = (the last char of remotePath="\")
	if is_dir then
		-- create a new local folder
		put "\"&$.file.getFileName(remotePath.char[1..remotePath.length-1]) after localDir
		ok = $.filesystem.folderCreate(localDir)

		rows = []
		res = $.wfx.FsFindFirst(remotePath)
		repeat while (res<>0)
			if res["FileName"] <>".." then rows.add(res)
			res = $.wfx.FsFindNext()
		end repeat
		$.wfx.FsFindClose()

		repeat with row in rows
			fn = row["FileName"]
			if (row["FileAttributes"] contains "D") then put "\" after fn
			me._saveLinuxPath(remotePath&fn, localDir)
		end repeat

	else
		-- "FsGetFile object me, string remotePath, string localPath, *flags\n"
		res = $.wfx.FsGetFile(remotePath, localDir&"\"&$.file.getFileName(remotePath))
	end if
end

on __SLOTS_CREATE
end

--------------------------------------
-- @callback
--------------------------------------
on slotQuit (me, action)
	$.shell.shell_cmd_thread_stop()
	me.slotMountUnmountAll()
	me.pState.setValue("LastDir", me.pLastDir)
	-- save window state
	val = me.pMainWindow.saveGeometry()
	me.pState.setValue("MainWindow/Geometry", val)
	val = me.pMainWindow.saveState()
	me.pState.setValue("MainWindow/State", val)
	_movie.halt()
end

--------------------------------------
-- @callback
--------------------------------------
on slotAbout (me)
	msg = "<b>"&me.APP_NAME&&"v0."&me.APP_VERSION&"</b>"&$.file.getString($.@("about.htm"))
	$.Qt.MessageBox.about(me.pMainWindow, "About "&me.APP_NAME, msg)
end

--------------------------------------
-- @callback
--------------------------------------
on slotUpdateDisks (me)
	me.pDiskInfos = [:]

	me.pComboBoxCreateDisk.clear()
	me.pComboBoxRestoreDisk.clear()

	-- test: add CD drives
	cmd = "ls /dev/scd*"
	res = $.shell.shell_cmd(cmd, ["return_list":TRUE]) -- ["/dev/scd0"]
	if res.count then
		res.sort()
		cnt = res.count
		repeat with i = 1 to cnt
			str = "CD " & (i-1)
			diskName = res[i].char[6..res[i].length]
			-- get drive letter
			info = $.shell.shell_cmd("cygpath -w "&res[i]) -- "\\.\F:
			put " ["&info.char[5]&":\]" after str
			-- get size
			info = $.shell.shell_cmd("fdisk -l "&res[i])
			put " - "&me._formatFileSize(info.word[5]) after str
			me.pComboBoxCreateDisk.addItem(str, diskName)
		end repeat
		me.pComboBoxCreateDisk.insertSeparator(cnt)
		--setItemEnabled (int index, bool on)
	end if

	-- note: "cat /proc/partitions" doesn't show mounted dokan volumes, but "mount" and "df" do
	cmd = "cat /proc/partitions"
	res = $.shell.shell_cmd(cmd, ["return_list":TRUE])
	if res.count=0 then return
	res.deleteAt(1)
	res.sort()

	repeat with row in res
		dev = row.word[4]
		if dev.length=3 then
			me.pDiskInfos[dev] = [#size:float(row.word[3])*1024, #mounts:[]]
		else if dev.length>3 and row.word.count>4 then
			me.pDiskInfos[dev.char[1..3]].mounts.add(row.word[5])
		end if
	end repeat
	--put me.pDiskInfos

	cnt = me.pDiskInfos.count
	repeat with i = 1 to cnt
		diskName = me.pDiskInfos.getPropAt(i)
		disk = me.pDiskInfos[i]
		str = "Disk " & (i-1)
		if disk.mounts.count then
			put " ["&$.string.implode(", ", disk.mounts)&"]" after str
		end if
		put " - "&me._formatFileSize(disk.size) after str
		me.pComboBoxCreateDisk.addItem(str, diskName)
		me.pComboBoxRestoreDisk.addItem(str, diskName)
	end repeat

	-- for safety reasons we select the last entry by default, which is more likely a removable drive
	n = me.pComboBoxCreateDisk.count-1
	me.pComboBoxCreateDisk.setCurrentIndex(n)
	me.slotCreateShowDiskInfo(me.pComboBoxCreateDisk, n)
	me.pComboBoxRestoreDisk.setCurrentIndex(n)
	me.slotRestoreShowDiskInfo(me.pComboBoxRestoreDisk, n)
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateFileSelect (me)
	if me.pComboBoxCreateDisk.currentText() starts "CD " then
		filter = "ISO Image Files (*.iso)"
	else
		filter = "Raw Disk Image Files (*.dd *.dsk *.img);;Compressed Disk Image Files (*.img.gz);;All Files (*)"
	end if
	fn = $.Qt.FileDialog.getSaveFileName(me.pMainWindow, "Image File", me.pLastDir, filter)
	if fn="" then return FALSE
	fn = $.@(fn)
	me.pLastDir = $.file.getFilePath(fn)
	me.pLineEditCreateFile.setText(fn)
	me.pPushButtonCreateStart.setDisabled(FALSE)
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateShowDiskInfo (me, widget, index)
	disk = widget.itemData(index)
	cmd = "fdisk -l /dev/"&disk & " 2>&1"
	info = $.shell.shell_cmd(cmd)
	info = info.word[1..info.word.count]
	lines = $.string.explodeBySingleChar($.LF, info)
	me.pCreateDiskSize = float(lines[1].word[5])
	me.pCreateLastSector = 0
	repeat with l in lines
		if l starts "Sector size" then
			--Sector size (logical/physical): 2048 bytes / 2048 bytes
			me.pCreateSectorSize = integer(l.word[4])
		end if
		if l starts "Disklabel type" then
			me.pCreateDiskIsGPT = (l.word[3]="gpt")
			me.pCheckBoxCreateOnlyPartitions.setDisabled(me.pCreateDiskIsGPT)
		end if
		if l starts "/dev" then
			-- Device     Boot Start       End   Sectors   Size Id Type
			-- /dev/sda1  *     2048 468862127 468860080 223.6G  7 HPFS/NTFS/exFAT
			endSector = float(l.word[4])
			if endSector>me.pCreateLastSector then me.pCreateLastSector=endSector
		end if
	end repeat

	if disk starts "scd" then
		cmd = "isoinfo -d -i /dev/"&disk & " 2>&1"
		res = $.shell.shell_cmd(cmd)
		put $.LF&$.LF & res.word[1..res.word.count] after info
	end if

	me.pPlainTextEditCreateDiskInfo.setPlainText(info)
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateStart (me, skipZero)
	disk = me.pComboBoxCreateDisk.itemData(me.pComboBoxCreateDisk.currentIndex())
	fn = me.pLineEditCreateFile.text()

	fType = $.file.getFileType(fn)
	if fType="iso" then
		cmd = "start /b /wait /high dd bs="&me.pCreateSectorSize&" status=progress if=/dev/"&disk
		put cmd
		use_compressed = FALSE
		me.pCreateImgSize = me.pCreateDiskSize
	else
		use_compressed = (fType="gz")
		if skipZero<>TRUE and use_compressed and me.pDiskInfos[disk].mounts.count then
			tButtons = $.Qt.QMessageBox.Yes+$.Qt.QMessageBox.No
			msg = "You selected a compressed image as target. "&\
			"Do you want to optimize compression by first zeroing all free space on the win volume(s) of the selected disk? This might take a moment, but results in a smaller image file."
			res = $.Qt.MessageBox.question(me.pMainWindow, "Zero free space now?", msg, tButtons)
			if res = $.Qt.QMessageBox.Yes then
				me._uiCreateSetEnabled(FALSE)
				return me._zeroMemory(disk)
			end if
		end if

		-- unmount volumes
		me.pDiskInfos[disk][#volumes] = [:]
		repeat with m in me.pDiskInfos[disk].mounts
			me.pDiskInfos[disk][#volumes][m] = $.shell.shell_cmd("mountvol "&m&" /L").word[1]
			$.shell.shell_cmd("mountvol "&m&" /P")
		end repeat

		if me.pCheckBoxCreateOnlyPartitions.isChecked() and not me.pCreateDiskIsGPT then
			me.pCreateImgSize = (me.pCreateLastSector+1)*me.pCreateSectorSize
			cmd = "start /b /wait /high dd count="&me._int(me.pCreateImgSize)&" iflag=count_bytes bs=1M status=progress if=/dev/"&disk
		else
			me.pCreateImgSize = me.pCreateDiskSize
			cmd = "start /b /wait /high dd bs=1M status=progress if=/dev/"&disk
		end if

	end if

	-- check free space on selected disk
	if not use_compressed and fn.char[2]=":" then
		letter = fn.char[1]
		res = $.shell.shell_cmd("df /cygdrive/"&letter)
		kb = float(res.word[11])
		if floatP(kb) then
			if me.pCreateImgSize>kb*1024 then
				tButtons = $.Qt.QMessageBox.Yes+$.Qt.QMessageBox.No
				msg = "Size of image larger than free disk space:<br><br>"&\
				"Image: "&me._int(me.pCreateImgSize)&" bytes<br>"&\
				"Disk: "&me._int(kb*1024)&" bytes<br><br>"&\
				"Continue anyway?"
				res = $.Qt.MessageBox.warning(me.pMainWindow, "Not enough disk space!", msg, tButtons)
				if res <> $.Qt.QMessageBox.Yes then return
			end if
		end if
	end if

	if use_compressed then
		put " | gzip >"&QUOTE&fn&QUOTE after cmd
		props = [:]
		props["cb_stderr"] = #slotCreateErr
		props["cb_complete"] = #slotCreateComplete
		props["cb_target"] = me
		props["timeout"] = 0
		$.shell.shell_cmd_thread(cmd, props)
	else
		put " of="&QUOTE&fn&QUOTE after cmd
		props = [:]
		props["cb_stderr"] = #slotCreateErr
		props["cb_complete"] = #slotCreateComplete
		props["cb_target"] = me
		props["timeout"] = 0
		$.shell.shell_cmd_thread(cmd, props)
	end if

	me.pProgressBar.setValue(0)

	taskBarProgressSetState(2) -- TBPF_NORMAL
	taskBarProgressSetValue(0)

	me.pProgressBar.show()
	me._uiCreateSetEnabled(FALSE)
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateCancel (me)
	$.shell.shell_cmd_thread_stop()
	me._uiCreateSetEnabled(TRUE)
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateErr (me, res)
	bytes = float(res.word[1])
	if floatP(bytes) then
		prog = bytes/me.pCreateImgSize
		me.pProgressBar.setValue(integer(100*prog))
		taskBarProgressSetValue(integer(100*prog))
		me.pStatusBar.showMessage("Creating image... ("&res.word[res.word.count-1..res.word.count]&")")
	end if
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateComplete (me, res)
	me._uiCreateSetEnabled(TRUE)
	taskBarProgressSetState(0) -- TBPF_NOPROGRESS
	if res=EXIT_CODE_ABORT then
		me.pStatusBar.showMessage("Running task was aborted by user.")
	else if res<>0 then
		put res
		me.pStatusBar.showMessage("Failure! Exit code: "&res)
		me.pTrayIcon.showMessage("Failure", "Creating image file failed with exit code"&&res&".", 3, 20000)
	else
		me.pStatusBar.showMessage("Done.")
		me.pTrayIcon.showMessage("Done", "Device was successfully saved to image file.", 1, 20000)
		me.pPushButtonCreateHash.setDisabled(FALSE)
	end if

	-- re-mount volumes
	disk = me.pComboBoxCreateDisk.itemData(me.pComboBoxCreateDisk.currentIndex())
	if voidP(me.pDiskInfos[disk]) then return
	repeat with i = 1 to me.pDiskInfos[disk].volumes.count
		m = me.pDiskInfos[disk].volumes.getPropAt(i)
		$.shell.shell_cmd("mountvol "&m&&me.pDiskInfos[disk].volumes[i])
	end repeat
end

--------------------------------------
-- @callback
--------------------------------------
on slotCreateGenerateHash (me)
	hashType = me.pComboBoxCreateHash.currentText()
	me.pStatusBar.showMessage("Calculating "&hashType&" Hash...")
	fn = me.pLineEditCreateFile.text()
	case hashType of
	"MD5":
		hash = $.crypto.cx_md5_file(fn)
	"SHA-1":
		hash = $.crypto.cx_sha1_file(fn)
	"SHA-256":
		hash = $.crypto.cx_sha256_file(fn)
	end case
	me.pLineEditCreateHash.setText(hash)
	me.pStatusBar.showMessage("Done.")
end

on __SLOTS_RESTORE
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreFileSelect (me)
	filter = "Disk Image Files (*.dd *.dsk *.img *.img.gz);;All Files (*)"
	fn = $.Qt.FileDialog.getOpenFileName(me.pMainWindow, "Image File", me.pLastDir, filter)
	if fn="" then return FALSE

	fn = $.@(fn)
	me.pLastDir = $.file.getFilePath(fn)

	me.pLineEditRestoreFile.setText(fn)

	me.pPushButtonRestoreStart.setDisabled(FALSE)
	me.pPushButtonRestoreHash.setDisabled(FALSE)

	-- try to get infos
	is_compressed = $.file.getFileType(fn)="gz"
	if is_compressed then
		me.pCheckBoxRestoreOnlyPartitions.setDisabled(TRUE)
		info = "No filesystem/partition info available for compressed images."
	else
		me.pCheckBoxRestoreOnlyPartitions.setDisabled(FALSE)
		cmd = "fdisk -l "&QUOTE&fn&QUOTE & "2>&1"
		info = $.shell.shell_cmd(cmd)
		info = info.word[1..info.word.count]

		lines = $.string.explodeBySingleChar($.LF, info)
		me.pRestoreLastSector = 0
		repeat with l in lines
			if l starts "Sector size" then
				--Sector size (logical/physical): 2048 bytes / 2048 bytes
				me.pRestoreSectorSize = integer(l.word[4])
			end if
			if l starts "Disklabel type" then
				me.pRestoreImageIsGPT = (l.word[3]="gpt")
				me.pCheckBoxRestoreOnlyPartitions.setDisabled(me.pRestoreImageIsGPT)
			end if
			if l starts fn then
				-- Device                    Boot Start     End Sectors  Size Id Type
				-- C:\TMP\Debian_QLingo.img1 *     2048 8060415 8058368  3.9G  c W95 FAT32 (LBA)
				delete char 1 to fn.length of l
				endSector = float(l.word[4])
				if endSector>me.pRestoreLastSector then me.pRestoreLastSector=endSector
			end if
		end repeat
	end if
	me.pPushButtonRestoreHash.setDisabled(is_compressed)
	me.pPlainTextEditRestoreImageInfo.setPlainText(info)
	me.pPlainTextEditRestoreImageInfo.moveCursorPosition(11)
	me.pPlainTextEditRestoreImageInfo.ensureCursorVisible()
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreShowDiskInfo (me, widget, index)
	disk = widget.itemData(index)

	cmd = "fdisk -l /dev/"&disk & " 2>&1"
	info = $.shell.shell_cmd(cmd)
	info = info.word[1..info.word.count]

	me.pRestoreDiskSize = float(info.word[5])

	me.pPlainTextEditRestoreDiskInfo.setPlainText(info)
	me.pPlainTextEditRestoreDiskInfo.moveCursorPosition(11)
	me.pPlainTextEditRestoreDiskInfo.ensureCursorVisible()
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreStart (me)
	index = me.pComboBoxRestoreDisk.currentIndex()
	disk = me.pComboBoxRestoreDisk.itemData(index)
	diskName = me.pComboBoxRestoreDisk.itemText(index)

	-- show confirm dialog
	tButtons = $.Qt.QMessageBox.Yes+$.Qt.QMessageBox.No
	res = $.Qt.MessageBox.warning(me.pMainWindow, APP_NAME, "All contents of "&diskName&" will be erased!<br><br>"&\
	"Do you really want to continue?", tButtons)
	if res <> $.Qt.QMessageBox.Yes then return

	fn = me.pLineEditRestoreFile.text()

	is_compressed = ($.file.getFileType(fn)="gz")
	if is_compressed then
		-- find uncompress filesize - ONLY works for uncompressed sizes < 4 GB!
--		res = $.shell.shell_cmd("gzip -l "&QUOTE&fn&QUOTE)
--    	me.pRestoreImgSize = float(res.word[6])
    	me.pRestoreImgSize = 0
	else
		if me.pCheckBoxRestoreOnlyPartitions.isChecked() and not me.pRestoreImageIsGPT then
			me.pRestoreImgSize = (me.pRestoreLastSector+1)*me.pRestoreSectorSize
		else
			me.pRestoreImgSize = $.filesystem.fileSize(fn)
		end if
	end if

	-- check sizes
	if me.pRestoreImgSize>me.pRestoreDiskSize then
		tButtons = $.Qt.QMessageBox.Yes+$.Qt.QMessageBox.No
		msg = "Size of image larger than device:<br><br>"&\
		"Image: "&me._int(me.pRestoreImgSize)&" bytes<br>"&\
		"Device: "&me._int(me.pRestoreDiskSize)&" bytes<br><br>"&\
		"Continue anyway?"
		res = $.Qt.MessageBox.warning(me.pMainWindow, "Size mismatch!", msg, tButtons)
		if res <> $.Qt.QMessageBox.Yes then return
		me.pRestoreImgSize = me.pRestoreDiskSize

		if is_compressed then
			--cmd = "gzip -d -c "&QUOTE&fn&QUOTE&" | start /b /wait /high dd count="&me._int(me.pRestoreImgSize)&" iflag=count_bytes bs=1M status=progress of=/dev/"&disk
			cmd = "gzip -d -c "&QUOTE&fn&QUOTE&" | start /b /wait /high dd iflag=count_bytes bs=1M status=progress of=/dev/"&disk
		else
			cmd = "start /b /wait /high dd count="&me._int(me.pRestoreImgSize)&" iflag=count_bytes bs=1M status=progress if="&QUOTE&fn&QUOTE&" of=/dev/"&disk
		end if
	else
		if is_compressed then
			cmd = "gzip -d -c "&QUOTE&fn&QUOTE&" | start /b /wait /high dd bs=1M status=progress of=/dev/"&disk
		else
			if me.pCheckBoxRestoreOnlyPartitions.isChecked() and not me.pRestoreImageIsGPT then
				cmd = "start /b /wait /high dd count="&me._int(me.pRestoreImgSize)&" iflag=count_bytes bs=1M status=progress if="&QUOTE&fn&QUOTE&" of=/dev/"&disk
			else
				cmd = "start /b /wait /high dd bs=1M status=progress if="&QUOTE&fn&QUOTE&" of=/dev/"&disk
			end if
		end if
	end if

	-- unmount volumes? -> "dd: error writing '/dev/sdd': No medium found"

	props = [:]
	props["cb_stderr"] = #slotRestoreErr
	props["cb_complete"] = #slotRestoreComplete
	props["cb_target"] = me
	props["timeout"] = 0
	$.shell.shell_cmd_thread(cmd, props)

	if me.pRestoreImgSize>0 then
		me.pProgressBar.setValue(0)
		taskBarProgressSetState(2) -- 2 = TBPF_NORMAL
		taskBarProgressSetValue(0)
		me.pProgressBar.show()
	else
		--taskBarProgressSetState(0) -- TBPF_NOPROGRESS
		taskBarProgressSetState(1) -- 1 = TBPF_INDETERMINATE = indicator cycles repeatedly
		me.pStatusBar.showMessage("Restoring from image...")
	end if

	me._uiRestoreSetEnabled(FALSE)
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreCancel (me)
	$.shell.shell_cmd_thread_stop()
	me._uiRestoreSetEnabled(TRUE)
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreErr (me, res)
	res = res.word[1..res.word.count]
	bytes = float(res.word[1])
	if floatP(bytes) then
		if me.pRestoreImgSize>0 then
			prog = bytes/me.pRestoreImgSize
			me.pProgressBar.setValue(integer(100*prog))
			taskBarProgressSetValue(integer(100*prog))
			me.pStatusBar.showMessage("Restoring from image... ("&res.word[res.word.count-1..res.word.count]&")")
		else
			me.pStatusBar.showMessage("Restoring from image... ("&me._formatFileSize(bytes)&" written)")
		end if
	end if
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreComplete (me, res)
	me._uiRestoreSetEnabled(TRUE)
	taskBarProgressSetState(0) -- TBPF_NOPROGRESS
	if res=EXIT_CODE_ABORT then
		me.pStatusBar.showMessage("Running task was aborted by user.")
	else if res<>0 then
		me.pStatusBar.showMessage("Failure! Exit code: "&res)
		me.pTrayIcon.showMessage("Failure", "Writing image file to device failed with exit code"&&res&".", 3, 20000)
	else
		me.pStatusBar.showMessage("Done.")
		me.pTrayIcon.showMessage("Done", "Image file was successfully written to the device.", 1, 20000)
	end if
	me.slotUpdateDisks()
end

--------------------------------------
-- @callback
--------------------------------------
on slotRestoreGenerateHash (me)
	hashType = me.pComboBoxRestoreHash.currentText()
	me.pStatusBar.showMessage("Calculating "&hashType&" Hash...")
	fn = me.pLineEditRestoreFile.text()
	case hashType of
	"MD5":
		hash = $.crypto.cx_md5_file(fn)
	"SHA-1":
		hash = $.crypto.cx_sha1_file(fn)
	"SHA-256":
		hash = $.crypto.cx_sha256_file(fn)
	end case
	me.pLineEditRestoreHash.setText(hash)
	me.pStatusBar.showMessage("Done.")
end

--------------------------------------
-- @callback
--------------------------------------
on slotWindowStateChanged (me)
	if me.pMainWindow.isMinimized() then
		me.pTrayIcon.show()
		--me.pMainWindow.hide()
	end if
end

--------------------------------------
-- @callback
--------------------------------------
on slotTrayIconActivated (me)
	--me.pMainWindow.show()
	me.pMainWindow.setWindowState(8)
	me.pTrayIcon.hide()
end

on __SLOTS_MOUNT
end

--------------------------------------
-- @callback
--------------------------------------
on slotMountNew (me)
	filter = "Raw Disk Image Files (*.dd *.dsk *.img);;VHD Disk Image Files (*.vhd"
	if me.pIsWindows8Point1OrGreater then put " *.vhdx" after filter -- VHDX only supported by Win 8.1 and later
	put ");;ISO Image Files (*.iso);;All Files (*)" after filter
	fn_org = $.Qt.FileDialog.getOpenFileName(me.pMainWindow, "Image File", me.pLastDir, filter)
	if fn_org="" then return FALSE
	fn_org = $.@(fn_org)

	me.pLastDir = $.file.getFilePath(fn_org)

	fn = fn_org
	ftype = $.file.getFileType(fn_org)
	case ftype of

	"iso":
		--cmd = "powershell -Command Mount-DiskImage -ImagePath "&QUOTE&fn&QUOTE & "&echo %ERRORLEVEL%"
		cmd = "powershell -Command "&QUOTE&"Mount-DiskImage -ImagePath '"&fn&"' -PassThru | Get-Volume"&QUOTE&" &echo %ERRORLEVEL%"
		--exitCode = $.shell.shell_cmd(cmd).word[1]
		res = $.shell.shell_cmd(cmd)
		exitCode = res.word[res.word.count]
		--put "exitCode", exitCode
		if exitCode<>"0" then
			msg = "Mounting image file '"&fn&"' failed."
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
			return
		end if

	"vhd", "vhdx":
		cmd = "(echo select vdisk file="&QUOTE&fn&QUOTE&" & echo attach vdisk) | diskpart 1>nul &echo %ERRORLEVEL%"
		exitCode = $.shell.shell_cmd(cmd).word[1]
		--put "exitCode2", exitCode
		if exitCode<>"0" then
			msg = "Mounting image file '"&fn&"' failed."
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
			return
		end if

	otherwise: -- raw image
		fn = fn_org & ".vhd"
		ok = $.filesystem.fileRename(fn_org, fn)
		if not ok then
			msg = "Temporarily renaming image file '"&fn_org&"' failed - is it write-protected?"
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
			return
		end if

		--> appends 512 bytes to end!
		cmd = "vhdtool /convert "&fn&" 1>nul 2>nul &echo %ERRORLEVEL%"
		exitCode = integer($.shell.shell_cmd(cmd).word[1])
		--put "exitCode1", exitCode
		if exitCode<>0 then
			msg = "Temporarily appending a VHD footer to image file '"&fn&"' failed."
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
			$.filesystem.fileRename(fn, fn_org)
			return
		end if

		cmd = "(echo select vdisk file="&QUOTE&fn&QUOTE&" & echo attach vdisk) | diskpart 1>nul &echo %ERRORLEVEL%"
		exitCode = $.shell.shell_cmd(cmd).word[1]
		--put "exitCode2", exitCode
		if exitCode<>"0" then
			msg = "Mounting image file '"&fn&"' failed."
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)

			$.filesystem.fileRename(fn, fn_org)

			-- remove footer
			fsize = bx_file_size(fn_org)
			ok = bx_file_truncate(fn_org, fsize-512)
			return
		end if

	end case

	me.pStatusBar.showMessage("Image file successfully mounted.")

	row = me.pTableWidgetMount.rowCount()
	me.pTableWidgetMount.setRowCount(row+1)

	me.slotUpdateDisks()
	me.slotLinuxUpdatePartitions()

	-- disk
	tItem = $.Qt.TableWidgetItem.new()
	if fType="iso" then
		tItem.setText(res.word[16]) -- FileSystemLabel
	else
		tItem.setText("/dev/"&me.pDiskInfos.getPropAt(me.pDiskInfos.count))
	end if
	me.pTableWidgetMount.setItem(row, 0, tItem)

	-- image file
	tItem = $.Qt.TableWidgetItem.new()
	tItem.setText(fn)
	tItem.setData($.Qt.UserRole, fn_org)
	me.pTableWidgetMount.setItem(row, 1, tItem)

	-- mounts
	tItem = $.Qt.TableWidgetItem.new()
	if fType="iso" then
		tItem.setText(res.word[15]&":\") -- DriveLetter
	else
		tItem.setText($.string.implode(" ", me.pDiskInfos.getLast().mounts))
	end if
	me.pTableWidgetMount.setItem(row, 2, tItem)

	me.pPushButtonUnmountAll.setDisabled(FALSE)
end

--------------------------------------
-- @callback
--------------------------------------
on slotMountTableWidgetCurrentItemChanged (me, w, currentItem, previousItem)
	flag = (currentItem.text()=0)
	me.pPushButtonUnmount.setDisabled(flag)
end

--------------------------------------
-- @callback
--------------------------------------
on slotMountTableWidgetCellDoubleClicked (me, w, row, col)
	if col<>2 then return
	mounts = me.pTableWidgetMount.item_(row, col).text()
	if mounts<>"" then $.shell.shell_cmd("explorer "&mounts.word[1])
end

--------------------------------------
-- @callback
--------------------------------------
on slotMountUnmount (me)
	row =  me.pTableWidgetMount.currentRow()
	if row<0 then return

	tItem = me.pTableWidgetMount.item_(row, 1)
	fn = tItem.text()

	ftype = $.file.getFileType(fn)
	if fType="iso" then
		cmd = "powershell -Command Dismount-DiskImage -ImagePath "&QUOTE&fn&QUOTE & "&echo %ERRORLEVEL%"
		exitCode = $.shell.shell_cmd(cmd).word[1]
		--put "exitCode", exitCode
		if exitCode<>"0" then
			msg = "Unmounting image file '"&fn&"' failed."
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
			return
		end if
	else
		fn_org = tItem.data($.Qt.UserRole)

		--cmd = "UnMountVHD "&fn&"  1>nul 2>nul &echo %ERRORLEVEL%"
		cmd = "(echo select vdisk file="&QUOTE&fn&QUOTE&" & echo detach vdisk) | diskpart 1>nul &echo %ERRORLEVEL%"
		exitCode = $.shell.shell_cmd(cmd).word[1]
		--put "exitCode", exitCode
		if exitCode<>"0" then
			msg = "Unmounting image file '"&fn&"' failed."
			$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
			return
		end if

		if fn_org<>fn then

			ok = $.filesystem.fileRename(fn, fn_org)
			if not ok then
				msg = "Renaming image file '"&fn&"' to its original name failed."
				$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
				return
			end if

			-- remove footer
			fsize = bx_file_size(fn_org)
			ok = bx_file_truncate(fn_org, fsize-512)
			if not ok then
				msg = "Removing temporary VHD footer from image file '"&fn_org&"' failed."
				$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
				return
			end if
		end if
	end if

	me.pStatusBar.showMessage("Image file successfully unmounted.")

	me.pTableWidgetMount.removeRow(row)
	me.pPushButtonUnmountAll.setDisabled(me.pTableWidgetMount.rowCount()=0)
	me.slotUpdateDisks()
	me.slotLinuxUpdatePartitions()
end

--------------------------------------
-- @callback
--------------------------------------
on slotMountUnmountAll (me)
	cnt = me.pTableWidgetMount.rowCount()
	if cnt then
		unmount_ok = 0

		repeat with i = cnt-1 down to 0

			tItem = me.pTableWidgetMount.item_(i, 1)
			fn = tItem.text()

			ftype = $.file.getFileType(fn)
			if fType="iso" then
				cmd = "powershell -Command Dismount-DiskImage -ImagePath "&QUOTE&fn&QUOTE & "&echo %ERRORLEVEL%"
				exitCode = $.shell.shell_cmd(cmd).word[1]
				--put "exitCode", exitCode
				if exitCode<>"0" then
					msg = "Unmounting image file '"&fn&"' failed."
					$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
					next repeat
				end if

			else
				fn_org = tItem.data($.Qt.UserRole)

				cmd = "(echo select vdisk file="&QUOTE&fn&QUOTE&" & echo detach vdisk) | diskpart 1>nul &echo %ERRORLEVEL%"
				exitCode = $.shell.shell_cmd(cmd).word[1]
				if exitCode<>"0" then
					msg = "Unmounting image file '"&fn&"' failed."
					$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
					next repeat
				end if

				if fn<>fn_org then
					ok = $.filesystem.fileRename(fn, fn_org)
					if not ok then
						msg = "Renaming image file '"&fn&"' to its original name failed."
						$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
						next repeat
					end if

					-- remove footer
					fsize = bx_file_size(fn_org)
					ok = bx_file_truncate(fn_org, fsize-512)
					if not ok then
						msg = "Removing temporary VHD footer from image file '"&fn_org&"' failed."
						$.Qt.MessageBox.critical(me.pMainWindow, "Failure", msg)
						return
					end if
				end if
			end if

			me.pTableWidgetMount.removeRow(i)
			unmount_ok = unmount_ok+1
		end repeat

		if unmount_ok=cnt then
			me.pStatusBar.showMessage("All image files successfully unmounted.")
			me.pPushButtonUnmountAll.setDisabled(TRUE)
		else if unmount_ok>0 then
			me.pStatusBar.showMessage("Not all image files were unmounted successfully.")
		else
			me.pStatusBar.showMessage("Failed to unmount image file(s).")
		end if
	end if

	me.slotUpdateDisks()
	me.slotLinuxUpdatePartitions()
end

on __SLOTS_LINUX
end

--------------------------------------
-- @callback
--------------------------------------
on slotLinuxUpdatePartitions (me)
	$.wfx.FsUnloadPlugin()
	$.wfx.FsLoadPlugin($.wfx.LIBDIR&"plugins\ext4Plugin\ext4Plugin.wfx")

	me.pComboBoxLinuxPartition.clear()
	me.pTableWidgetLinux.clearContents()

	tRootPath = "\"
	res = $.wfx.FsFindFirst(tRootPath)
	if res<>0 then
		me.pComboBoxLinuxPartition.addItem("/dev/"&res["FileName"], res["FileName"])
		repeat while (res<>0)
			res = $.wfx.FsFindNext()
			if res<>0 then
				me.pComboBoxLinuxPartition.addItem("/dev/"&res["FileName"], res["FileName"])
			end if
		end repeat
		$.wfx.FsFindClose()
	end if
	index = me.pComboBoxLinuxPartition.currentIndex()
	if index>=0 then me._loadLinuxPath("\"&me.pComboBoxLinuxPartition.itemData(index)&"\")
end

--------------------------------------
-- @callback
--------------------------------------
on slotTableWidgetLinuxCellDoubleClicked (me, w, row, col)
	tItem = me.pTableWidgetLinux.item_(row,0)
	p = tItem.data($.Qt.UserRole)
	if the last char of p="\" then
		me._loadLinuxPath(p)
	else
		me.slotTableWidgetLinuxSave()
	end if
end

--------------------------------------
-- @callback
--------------------------------------
on slotTableWidgetLinuxContextMenuRequested (me)
	currItem = me.pTableWidgetLinux.currentItem()
	if voidP(currItem) then return
	if currItem.text()=".." then return
	me.pTableWidgetLinux[#contextMenu].exec()
end

--------------------------------------
-- @callback
--------------------------------------
on slotTableWidgetLinuxSave (me)
	currItem = me.pTableWidgetLinux.item_(me.pTableWidgetLinux.currentRow(), 0)
	localDir = $.Qt.FileDialog.getExistingDirectory(me.pMainWindow, "Save to Folder", me.pLastDir, $.Qt.QFileDialog.ShowDirsOnly)
	if localDir="" then return
	localDir = $.@(localDir)
	me.pLastDir = localDir
	me.pStatusBar.showMessage("Saving files...")
	me.pProgressBar.setValue(0)
	me.pProgressBar.show()
	me.pTabWidget.setDisabled(TRUE)
	me._saveLinuxPath(currItem.data($.Qt.UserRole), localDir)
	me.pProgressBar.hide()
	me.pTabWidget.setDisabled(FALSE)
	me.pStatusBar.showMessage("Done.")
end

--------------------------------------
-- @callback
--------------------------------------
on slotLinuxGetFileProgress (me, pct)
	me.pProgressBar.setValue(pct)
end
